"""
RenderableCalculator tests.
"""

import pytest
from lute.models.term import Term
from lute.parse.base import ParsedToken
from lute.read.render.renderable_calculator import RenderableCalculator


def make_tokens(token_data):
    "Make ParsedTokens for the data."
    return [ParsedToken(t, 1) for t in token_data]


def assert_renderable_equals(
    language, token_data, term_data, expected, expected_displayed=None
):
    """
    Run the given scenario:

    Given tokens in language, and terms saved in that language,
    the calculator should render the expected terms,
    only displaying the appropriate parts of each.
    """
    tokens = make_tokens(token_data)
    terms = [Term(language, t) for t in term_data]

    rc = RenderableCalculator()
    rcs = rc.main(language, terms, tokens)
    res = ""
    for rc in rcs:
        res += f"[{rc.text}-{rc.length}]"

    zws = chr(0x200B)
    res = res.replace(zws, "")

    assert res == expected

    if expected_displayed is not None:
        res = ""
        for rc in rcs:
            res += f"[{rc.display_text}-{rc.length}]"

        res = res.replace(zws, "")
        assert res == expected_displayed


def test_simple_render(english):
    "Tokens with no defined terms are rendered as-is."
    data = ["some", " ", "data", " ", "here", "."]
    expected = "[some-1][ -1][data-1][ -1][here-1][.-1]"
    assert_renderable_equals(english, data, [], expected)


def test_non_matching_terms_are_ignored(english):
    "Non-match ignored."
    data = ["some", " ", "data", " ", "here", "."]
    expected = "[some-1][ -1][data-1][ -1][here-1][.-1]"
    assert_renderable_equals(english, data, ["ignoreme"], expected)


def test_partial_matching_terms_are_ignored(english):
    "Partial match is not the same as a match."
    data = ["some", " ", "data", " ", "here", "."]
    expected = "[some-1][ -1][data-1][ -1][here-1][.-1]"
    assert_renderable_equals(english, data, ["data he"], expected)


def test_tokens_must_be_contiguous(english):
    """
    If tokens aren't contiguous, the algorithm gets confused.
    """
    data = ["some", " ", "data", " ", "here", "."]
    tokens = make_tokens(data)
    tokens[1].order = 99
    rc = RenderableCalculator()
    with pytest.raises(Exception):
        rc.main(english, [], tokens)


def test_multiword_items_cover_other_items(english):
    """
    Given a multiword term, some of the other terms are hidden.
    """
    data = ["some", " ", "data", " ", "here", "."]
    words = [
        "data here",
    ]
    expected = "[some-1][ -1][data here-3][.-1]"
    assert_renderable_equals(english, data, words, expected)


def test_case_not_considered_for_matches(english):
    "Case doesnt matter."
    data = ["some", " ", "data", " ", "here", "."]
    expected = "[some-1][ -1][data here-3][.-1]"
    assert_renderable_equals(english, data, ["DATA HERE"], expected)


def test_term_found_in_multiple_places(english):
    "Term can be in a few places."
    data = [
        "some",
        " ",
        "data",
        " ",
        "here",
        " ",
        "more",
        " ",
        "data",
        " ",
        "here",
        ".",
    ]
    expected = "[some-1][ -1][data here-3][ -1][more-1][ -1][data here-3][.-1]"
    assert_renderable_equals(english, data, ["DATA HERE"], expected)


def test_overlapping_multiwords(english):
    """
    Given two overlapping terms, they're both displayed,
    but some of the second is cut off.
    """
    data = ["some", " ", "data", " ", "here", "."]
    words = [
        "some data",
        "data here",
    ]
    expected = "[some data-3][data here-3][.-1]"
    expected_displayed = "[some data-3][ here-3][.-1]"
    assert_renderable_equals(english, data, words, expected, expected_displayed)


def test_multiwords_starting_at_same_location(english):
    """
    Test overlapping matches.
    Given two terms that contain the same chars at the start,
    the longer term overwrites the shorter.
    """
    data = ["A", " ", "B", " ", "C", " ", "D"]
    words = [
        "A B",
        "A B C",
    ]
    expected = "[A B C-5][ -1][D-1]"
    assert_renderable_equals(english, data, words, expected)


def test_crazy_case(english):
    """
    Crazy test case covering the scenario in the class docstring.
    """
    data = [
        "A",
        " ",
        "B",
        " ",
        "C",
        " ",
        "D",
        " ",
        "E",
        " ",
        "F",
        " ",
        "G",
        " ",
        "H",
        " ",
        "I",
    ]
    words = [
        "B C",  # J
        "E F G H I",  # K
        "F G",  # L
        "C D E",  # M
    ]
    expected = "[A-1][ -1][B C-3][C D E-5][E F G H I-9]"
    expected_displayed = "[A-1][ -1][B C-3][ D E-5][ F G H I-9]"
    assert_renderable_equals(english, data, words, expected, expected_displayed)



"""
Utility methods for tests.
"""

from lute.models.term import Term
from lute.models.book import Book
from lute.read.render.service import get_paragraphs
from lute.db import db


def add_terms(language, term_array):
    """
    Make and save terms.
    """
    for term in term_array:
        t = Term(language, term)
        db.session.add(t)
    db.session.commit()


def make_book(title, content, language):
    """
    Make a book.
    """
    b = Book.create_book(title, language, content)
    return b


def make_text(title, content, language):
    """
    Make a single-page book, return the text.
    """
    b = make_book(title, content, language)
    return b.texts[0]


def get_rendered_string(text, imploder="/", overridestringize=None):
    "Get the stringized rendered content after parsing."

    def stringize(ti):
        zws = "\u200B"
        status = ""
        if ti.wo_status not in [None, 0]:
            status = f"({ti.wo_status})"
        return ti.display_text.replace(zws, "") + status

    usestringize = overridestringize or stringize
    ret = []
    paras = get_paragraphs(text.text, text.book.language)
    for p in paras:
        tis = [t for s in p for t in s.textitems]
        ss = [usestringize(ti) for ti in tis]
        ret.append(imploder.join(ss))
    return "/<PARA>/".join(ret)


def assert_rendered_text_equals(text, expected, msg=""):
    "Check that the rendered string matches the expected."
    actual = get_rendered_string(text)
    # This assertion gives details because the module
    # is registered in tests/__init__.py.
    assert actual == "/<PARA>/".join(expected), msg



"Smoke test only."

from lute.cli.language_term_export import generate_file

from lute.models.term import Term, TermTag
from lute.db import db


def test_smoke_test(app_context, tmp_path, english):
    "dump data."
    t = Term(english, "the")
    t.translation = "article"
    t.add_term_tag(TermTag("a"))
    t.add_term_tag(TermTag("b"))
    db.session.add(t)
    db.session.commit()

    outfile = tmp_path / "outfile.csv"
    generate_file("English", outfile)
    with open(outfile, "r", encoding="utf-8") as ofhandle:
        text = ofhandle.read()
    print(text)
    lines = text.split("\n")
    assert (
        lines[0] == "term,count,familycount,books,definition,status,children,tags"
    ), "headings"
    firstline = lines[1]
    assert firstline.startswith("the,"), "the is most common"
    assert firstline.endswith('article,1,-,"a, b"'), "ending data"



"""
Read service tests.
"""

from lute.models.term import Term
from lute.parse.base import ParsedToken
from lute.book.model import Book, Repository
from lute.read.render.service import find_all_Terms_in_string, get_paragraphs
from lute.read.service import start_reading
from lute.db import db

from tests.utils import add_terms, make_text, assert_rendered_text_equals
from tests.dbasserts import assert_record_count_equals, assert_sql_result


def _run_scenario(language, content, expected_found):
    """
    Given some pre-saved terms in language,
    find_all method returns the expected_found terms that
    exist in the content string.
    """
    found_terms = find_all_Terms_in_string(content, language)
    assert len(found_terms) == len(expected_found), "found count"
    zws = "\u200B"  # zero-width space
    found_terms = [t.text.replace(zws, "") for t in found_terms]
    assert found_terms is not None
    assert expected_found is not None
    found_terms.sort()
    expected_found.sort()
    assert found_terms == expected_found


def test_spanish_find_all_in_string(spanish, app_context):
    "Given various pre-saved terms, find_all returns those in the string."
    terms = ["perro", "gato", "un gato"]
    for term in terms:
        t = Term(spanish, term)
        db.session.add(t)
    db.session.commit()

    _run_scenario(spanish, "Hola tengo un gato", ["gato", "un gato"])
    _run_scenario(spanish, "gato gato gato", ["gato"])
    _run_scenario(spanish, "No tengo UN PERRO", ["perro"])
    _run_scenario(spanish, "Hola tengo un    gato", ["gato", "un gato"])
    _run_scenario(spanish, "No tengo nada", [])

    terms = ["échalo", "ábrela"]
    for term in terms:
        t = Term(spanish, term)
        db.session.add(t)
    db.session.commit()

    _run_scenario(spanish, '"Échalo", me dijo.', ["échalo"])
    _run_scenario(spanish, "gato ábrela Ábrela", ["gato", "ábrela"])


def test_english_find_all_in_string(english, app_context):
    "Can find a term with an apostrophe in string."
    terms = ["the cat's pyjamas"]
    for term in terms:
        t = Term(english, term)
        db.session.add(t)
    db.session.commit()

    _run_scenario(english, "This is the cat's pyjamas.", ["the cat's pyjamas"])


def test_turkish_find_all_in_string(turkish, app_context):
    "Finds terms, handling case conversion."
    terms = ["ışık", "için"]
    for term in terms:
        t = Term(turkish, term)
        db.session.add(t)
    db.session.commit()

    content = "Işık İçin."
    _run_scenario(turkish, content, ["ışık", "için"])


def test_smoke_get_paragraphs(spanish, app_context):
    """
    Smoke test to get paragraph information.
    """
    add_terms(spanish, ["tengo un", "un gato"])

    content = "Tengo un gato. Hay un perro.\nTengo un perro."
    t = make_text("Hola", content, spanish)
    db.session.add(t)
    db.session.commit()

    ParsedToken.reset_counters()
    paras = get_paragraphs(t.text, t.book.language)
    assert len(paras) == 2

    def stringize(t):
        zws = chr(0x200B)
        parts = [
            f"[{t.display_text.replace(zws, '/')}(",
            f"{t.para_id}.{t.se_id}",
            ")]",
        ]
        return "".join(parts)

    sentences = [item for sublist in paras for item in sublist]
    actual = []
    for sent in sentences:
        actual.append("".join(map(stringize, sent.textitems)))

    expected = [
        "[Tengo/ /un(0.0)][ /gato(0.0)][. (0.0)]",
        "[Hay(0.1)][ (0.1)][un(0.1)][ (0.1)][perro(0.1)][.(0.1)]",
        "[Tengo/ /un(1.3)][ (1.3)][perro(1.3)][.(1.3)]",
    ]
    assert actual == expected


def test_smoke_rendered(spanish, app_context):
    """
    Smoke test to get paragraph information.
    """
    add_terms(spanish, ["tengo un", "un gato"])
    content = ["Tengo un gato. Hay un perro.", "Tengo un perro."]
    text = make_text("Hola", "\n".join(content), spanish)
    db.session.add(text)
    db.session.commit()

    expected = ["Tengo un(1)/ gato(1)/. /Hay/ /un/ /perro/.", "Tengo un(1)/ /perro/."]
    assert_rendered_text_equals(text, expected)


def test_rendered_leaves_blank_lines(spanish, app_context):
    """
    Smoke test to get paragraph information.
    """
    add_terms(spanish, ["tengo un", "un gato"])
    content = ["Tengo un gato. Hay un perro.", "", "Tengo un perro."]
    text = make_text("Hola", "\n".join(content), spanish)
    db.session.add(text)
    db.session.commit()

    expected = [
        "Tengo un(1)/ gato(1)/. /Hay/ /un/ /perro/.",
        "",
        "Tengo un(1)/ /perro/.",
    ]
    assert_rendered_text_equals(text, expected)


## Start reading tests. ##########################


def test_smoke_start_reading(english, app_context):
    "Smoke test book."
    b = Book()
    b.title = "blah"
    b.language_id = english.id
    b.text = "Here is some content.  Here is more."
    r = Repository(db)
    dbbook = r.add(b)
    r.commit()

    assert_record_count_equals("select * from sentences", 0, "before start")
    start_reading(dbbook, 1, db.session)
    assert_record_count_equals("select * from sentences", 2, "after start")


def test_start_reading_creates_Terms_for_unknown_words(english, app_context):
    "Unknown (status 0) terms are created for all new words."
    t = Term(english, "dog")
    db.session.add(t)
    db.session.commit()

    b = Book()
    b.title = "blah"
    b.language_id = english.id
    b.text = "Dog CAT dog cat."
    r = Repository(db)
    dbbook = r.add(b)
    r.commit()

    sql = "select WoTextLC from words order by WoText"
    assert_sql_result(sql, ["dog"], "before start")

    paragraphs = start_reading(dbbook, 1, db.session)
    textitems = [
        ti
        for para in paragraphs
        for sentence in para
        for ti in sentence.textitems
        if ti.is_word and ti.wo_id is None
    ]
    assert (
        len(textitems) == 0
    ), f"All text items should have a term, but got {textitems}"
    assert_sql_result(sql, ["cat", "dog"], "after start")



"""
Testing management functions.

"management" is for global actions like clearing out the db.
"""

import pytest
from lute.db import db
from lute.models.setting import UserSetting, BackupSettings
from lute.db.management import delete_all_data
from tests.dbasserts import assert_record_count_equals


@pytest.mark.dbwipe
def test_wiping_db_clears_out_all_tables(app_context):
    """
    DB is wiped clean if requested ... settings are left!

    This test is also used from /tasks.py; see .pytest.ini.
    """
    old_user_settings = db.session.query(UserSetting).all()

    delete_all_data()
    tables = [
        "books",
        "bookstats",
        "booktags",
        "languages",
        "sentences",
        "tags",
        "tags2",
        "texts",
        "wordflashmessages",
        "wordimages",
        "wordparents",
        "words",
        "wordtags",
    ]
    for t in tables:
        assert_record_count_equals(t, 0, t)

    sql = "select * from settings where stkeytype='user'"
    assert_record_count_equals(sql, len(old_user_settings), "user settings remain")
    sql = "select * from settings where StKeyType = 'system'"
    assert_record_count_equals(sql, 0, "no system settings")


def test_can_get_backup_settings_when_db_is_wiped(app_context):
    "The backupsettings struct assumes certain things about the data."
    delete_all_data()
    bs = BackupSettings.get_backup_settings()
    assert bs.backup_enabled, "backup is back to being enabled"
    assert bs.backup_dir is not None, "default restored"



"""
SentenceGroupIterator tests.
"""

from lute.parse.base import SentenceGroupIterator
from lute.parse.space_delimited_parser import SpaceDelimitedParser


def toks_to_string(tokens):
    return "".join([t.token for t in tokens])


def test_sgi_scenarios(english):
    """
    Given a string and the max token count,
    SentenceGroupIterator should return expected groups.
    """

    def scenario(s, maxcount, expected_groups):
        parser = SpaceDelimitedParser()
        tokens = parser.get_parsed_tokens(s, english)

        it = SentenceGroupIterator(tokens, maxcount)
        groups = []
        while g := it.next():
            groups.append(g)
        gs = [toks_to_string(g) for g in groups]
        assert "||".join(gs) == "||".join(
            expected_groups
        ), f"groups for size {maxcount}"

    scenario("", 100, [""])

    text = "Here is a dog. Here is a cat."

    scenario(text, 100, ["Here is a dog. Here is a cat."])

    scenario(text, 3, ["Here is a dog. ", "Here is a cat."])

    scenario(text, 6, ["Here is a dog. ", "Here is a cat."])

    # No period at the end.
    scenario("Here is a dog. Here is a cat", 6, ["Here is a dog. ", "Here is a cat"])

    # No period at all.
    scenario("Here is a dog Here is a cat", 6, ["Here is a dog Here is a cat"])

    scenario(
        "Here is a dog. Here is a cat. Here is a thing.",
        10,
        ["Here is a dog. Here is a cat. ", "Here is a thing."],
    )



"""
Theme service tests.
"""

import os
import lute.themes.service as svc
from lute.db import db
from lute.models.setting import UserSetting


def test_list_themes(app_context):
    "Smoke test only."
    lst = svc.list_themes()
    assert len(lst) > 0, "have themes"
    assert lst[0][0] == "-", "No theme"
    assert lst[0][1] == "(default)"

    assert ("Apple_Books.css", "Apple Books") in lst


def test_default_theme_is_blank_css(app_context):
    "UserSetting starts off with blank css."
    assert UserSetting.get_value("current_theme") == "-"
    assert svc.get_current_css() == "", "Default = empty string."


def test_bad_setting_returns_blank_css(app_context):
    "Just in case."
    UserSetting.set_value("current_theme", "_missing_file.css")
    db.session.commit()
    assert svc.get_current_css() == "", "Missing = empty string."


def test_setting_a_theme_returns_its_css(app_context):
    "User choice is used."
    UserSetting.set_value("current_theme", "Apple_Books.css")
    db.session.commit()
    assert "Georgia" in svc.get_current_css(), "font specified"


def test_next_theme_cycles_themes(app_context):
    """
    Users should be able to move the 'next' theme quickly
    while reading, via a hotkey.
    """
    lst = svc.list_themes()
    assert UserSetting.get_value("current_theme") == lst[0][0]
    svc.next_theme()
    assert UserSetting.get_value("current_theme") == lst[1][0]
    for _ in range(0, len(lst) + 10):  # pylint: disable=consider-using-enumerate
        svc.next_theme()
        svc.next_theme()
    # OK


def _delete_custom_theme_files(theme_dir):
    "Delete custom file."
    for filename in os.listdir(theme_dir):
        filepath = os.path.join(theme_dir, filename)
        if os.path.isfile(filepath):
            os.remove(filepath)


def test_custom_theme_in_theme_dir_is_available(app, app_context):
    "Can use .css file in theme dir."
    theme_dir = app.env_config.userthemespath
    _delete_custom_theme_files(theme_dir)

    mytheme_content = "p { font-size: 30pt; }"
    themefile = os.path.join(theme_dir, "my_theme.css")
    with open(themefile, "w", encoding="utf-8") as f:
        f.write(mytheme_content)

    lst = svc.list_themes()
    assert ("my_theme.css", "my theme") in lst, "Have my theme"

    UserSetting.set_value("current_theme", "my_theme.css")
    db.session.commit()
    assert mytheme_content in svc.get_current_css(), "my theme used"


def test_custom_theme_in_theme_dir_appends_to_existing_theme(app, app_context):
    "Can use .css file in theme dir."
    theme_dir = app.env_config.userthemespath
    _delete_custom_theme_files(theme_dir)

    lst = svc.list_themes()
    assert ("Apple_Books.css", "Apple Books") in lst
    UserSetting.set_value("current_theme", "Apple_Books.css")
    db.session.commit()
    old_content = svc.get_current_css()

    mytheme_content = "p { font-size: 30pt; }"
    themefile = os.path.join(theme_dir, "Apple_Books.css")
    with open(themefile, "w", encoding="utf-8") as f:
        f.write(mytheme_content)

    lst = svc.list_themes()
    assert ("Apple_Books.css", "Apple Books") in lst, "Have my theme"

    UserSetting.set_value("current_theme", "Apple_Books.css")
    db.session.commit()

    new_css = old_content + "\n\n/* Additional user css */\n\n" + mytheme_content
    new_content = svc.get_current_css()
    assert new_css in new_content, "my theme used in addition to built-in"



"""
Data cleanup routines.

Sometimes required as data management changes.
These cleanup routines will be called by the app_factory.
"""

from lute.db import db
from lute.models.book import Text


def _set_texts_word_count():
    """
    texts.TxWordCount should be set for all texts.

    Fixing a design error: the counts should have been stored here,
    instead of only in books.BkWordCount.

    Ref https://github.com/jzohrab/lute-v3/issues/95
    """
    calc_counts = db.session.query(Text).filter(Text.word_count.is_(None)).all()

    # Don't recalc with invalid parsers!!!!
    recalc = [t for t in calc_counts if t.book.language.is_supported]

    if len(recalc) == 0:
        # Nothing to calculate, quit.
        return

    for t in recalc:
        pt = t.book.language.get_parsed_tokens(t.text)
        words = [w for w in pt if w.is_word]
        t.word_count = len(words)
        db.session.add(t)
    db.session.commit()


def clean_data():
    "Clean all data as required."
    _set_texts_word_count()



"""
/book routes.
"""

import json
from flask import (
    Blueprint,
    request,
    jsonify,
    render_template,
    redirect,
    flash,
)
from lute.utils.data_tables import DataTablesFlaskParamParser
from lute.book import service
from lute.book.datatables import get_data_tables_list
from lute.book.forms import NewBookForm, EditBookForm
import lute.utils.formutils
from lute.db import db

from lute.models.language import Language
from lute.models.book import Book as DBBook
from lute.models.setting import UserSetting
from lute.book.model import Book, Repository


bp = Blueprint("book", __name__, url_prefix="/book")


def _load_term_custom_filters(request_form, parameters):
    "Manually add filters that the DataTablesFlaskParamParser doesn't know about."
    filter_param_names = [
        "filtLanguage",
    ]
    request_params = request_form.to_dict(flat=True)
    for p in filter_param_names:
        parameters[p] = request_params.get(p)


def datatables_source(is_archived):
    "Get datatables json for books."
    parameters = DataTablesFlaskParamParser.parse_params(request.form)
    _load_term_custom_filters(request.form, parameters)
    data = get_data_tables_list(parameters, is_archived)
    return jsonify(data)


@bp.route("/datatables/active", methods=["POST"])
def datatables_active_source():
    "Datatables data for active books."
    return datatables_source(False)


@bp.route("/archived", methods=["GET"])
def archived():
    "List archived books."
    language_choices = lute.utils.formutils.language_choices("(all languages)")
    current_language_id = lute.utils.formutils.valid_current_language_id()

    return render_template(
        "book/index.html",
        status="Archived",
        language_choices=language_choices,
        current_language_id=current_language_id,
    )


# Archived must be capitalized, or the ajax call 404's.
@bp.route("/datatables/Archived", methods=["POST"])
def datatables_archived_source():
    "Datatables data for archived books."
    return datatables_source(True)


def _book_from_url(url):
    "Create a new book, or flash an error if can't parse."
    b = Book()
    try:
        b = service.book_from_url(url)
    except service.BookImportException as e:
        flash(e.message, "notice")
        b = Book()
    return b


def _language_is_rtl_map():
    """
    Return language-id to is_rtl map, to be used during book creation.
    """
    ret = {}
    for lang in db.session.query(Language).all():
        ret[lang.id] = lang.right_to_left
    return ret


@bp.route("/new", methods=["GET", "POST"])
def new():
    "Create a new book, either from text or from a file."
    b = Book()
    import_url = request.args.get("importurl", "").strip()
    if import_url != "":
        b = _book_from_url(import_url)

    form = NewBookForm(obj=b)
    form.language_id.choices = lute.utils.formutils.language_choices()
    repo = Repository(db)

    if form.validate_on_submit():
        try:
            form.populate_obj(b)
            book = repo.add(b)
            repo.commit()
            return redirect(f"/read/{book.id}/page/1", 302)
        except service.BookImportException as e:
            flash(e.message, "notice")

    # Don't set the current language before submit.
    current_language_id = int(UserSetting.get_value("current_language_id"))
    form.language_id.data = current_language_id

    return render_template(
        "book/create_new.html",
        book=b,
        form=form,
        tags=repo.get_book_tags(),
        rtl_map=json.dumps(_language_is_rtl_map()),
        show_language_selector=True,
    )


@bp.route("/edit/<int:bookid>", methods=["GET", "POST"])
def edit(bookid):
    "Edit a book - can only change a few fields."
    repo = Repository(db)
    b = repo.load(bookid)
    form = EditBookForm(obj=b)

    if form.validate_on_submit():
        form.populate_obj(b)
        repo.add(b)
        repo.commit()
        flash(f"{b.title} updated.")
        return redirect("/", 302)

    lang = Language.find(b.language_id)
    return render_template(
        "book/edit.html",
        book=b,
        title_direction="rtl" if lang.right_to_left else "ltr",
        form=form,
        tags=repo.get_book_tags(),
    )


@bp.route("/import_webpage", methods=["GET", "POST"])
def import_webpage():
    return render_template("book/import_webpage.html")


@bp.route("/archive/<int:bookid>", methods=["POST"])
def archive(bookid):
    "Archive a book."
    b = DBBook.find(bookid)
    b.archived = True
    db.session.add(b)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/unarchive/<int:bookid>", methods=["POST"])
def unarchive(bookid):
    "Archive a book."
    b = DBBook.find(bookid)
    b.archived = False
    db.session.add(b)
    db.session.commit()
    return redirect("/", 302)


@bp.route("/delete/<int:bookid>", methods=["POST"])
def delete(bookid):
    "Archive a book."
    b = DBBook.find(bookid)
    db.session.delete(b)
    db.session.commit()
    return redirect("/", 302)



"""
Tests for managing the demo data.
"""

from sqlalchemy import text
import pytest
from lute.db import db
from lute.db.demo import (
    contains_demo_data,
    remove_flag,
    delete_demo_data,
    tutorial_book_id,
    load_demo_data,
)
import lute.parse.registry
from tests.dbasserts import assert_record_count_equals, assert_sql_result


def test_new_db_is_demo(app_context):
    "New db created from the baseline has the demo flag set."
    assert contains_demo_data() is True, "new db contains demo."


def test_removing_flag_means_not_demo(app_context):
    "Unsetting the flag means the db is not a demo."
    remove_flag()
    assert contains_demo_data() is False, "not a demo."


def test_wiping_db_clears_flag(app_context):
    "No longer a demo if the demo is wiped out!"
    delete_demo_data()
    assert contains_demo_data() is False, "not a demo."


def test_wipe_db_only_works_if_flag_is_set(app_context):
    "Can only wipe a demo db!"
    remove_flag()
    with pytest.raises(Exception):
        delete_demo_data()


def test_tutorial_id_returned_if_present(app_context):
    "Sanity check."
    assert tutorial_book_id() > 0, "have tutorial"

    sql = 'update books set bktitle = "xxTutorial" where bktitle = "Tutorial"'
    db.session.execute(text(sql))
    db.session.commit()
    assert tutorial_book_id() is None, "no tutorial"

    sql = 'update books set bktitle = "Tutorial" where bktitle = "xxTutorial"'
    db.session.execute(text(sql))
    db.session.commit()
    assert tutorial_book_id() > 0, "have tutorial again"

    delete_demo_data()
    assert tutorial_book_id() is None, "no tutorial"


# Loading.


@pytest.mark.dbdemoload
def test_load_demo_loads_language_yaml_files(app_context):
    """
    All data is loaded, spot check some.

    This test is also used from "inv db.reset" in tasks.py
    (see .pytest.ini).
    """
    delete_demo_data()
    assert contains_demo_data() is False, "not a demo."
    assert_record_count_equals("languages", 0, "wiped out")

    # Wipe out all settings!!!
    # When user installs, the settings need to be loaded
    # with values from _their_ config and environment.
    sql = "delete from settings"
    db.session.execute(text(sql))
    db.session.commit()

    load_demo_data()
    assert contains_demo_data() is True, "demo loaded"
    checks = [
        "select * from languages where LgName = 'English'",
        "select * from books where BkTitle = 'Tutorial'",
    ]
    for c in checks:
        assert_record_count_equals(c, 1, c + " returned 1")

    sql = "select distinct stkeytype from settings"
    assert_sql_result(sql, ["system"], "only system settings remain")


@pytest.fixture(name="_restore_japanese_parser")
def fixture_restore_mecab_support():
    """
    "Teardown" method to restore jp parser if it was removed.
    """
    k = "japanese"
    assert k in lute.parse.registry.parsers, "have jp parser, sanity check"
    old_val = lute.parse.registry.parsers[k]

    yield

    if k not in lute.parse.registry.parsers:
        lute.parse.registry.parsers[k] = old_val



"""
Show terms in datatables.
"""

from lute.db import db
from lute.utils.data_tables import DataTablesSqliteQuery, supported_parser_type_criteria


def get_data_tables_list(parameters):
    "Term json data for datatables."

    base_sql = """SELECT
    0 as chk, w.WoID as WoID, LgName, L.LgID as LgID, w.WoText as WoText, parents.parentlist as ParentText, w.WoTranslation,
    w.WoRomanization,
    replace(wi.WiSource, '.jpeg', '') as WiSource,
    ifnull(tags.taglist, '') as TagList,
    StText,
    StID,
    StAbbreviation,
    case w.WoSyncStatus when 1 then 'y' else '' end as SyncStatus,
    datetime(WoCreated, 'localtime') as WoCreated
    FROM
    words w
    INNER JOIN languages L on L.LgID = w.WoLgID
    INNER JOIN statuses S on S.StID = w.WoStatus
    LEFT OUTER JOIN (
      SELECT WpWoID as WoID, GROUP_CONCAT(PText, ', ') AS parentlist
      FROM
      (
        select WpWoID, WoText as PText
        from wordparents wp
        INNER JOIN words on WoID = WpParentWoID
        order by WoText
      ) parentssrc
      GROUP BY WpWoID
    ) AS parents on parents.WoID = w.WoID
    LEFT OUTER JOIN (
      SELECT WtWoID as WoID, GROUP_CONCAT(TgText, ', ') AS taglist
      FROM
      (
        select WtWoID, TgText
        from wordtags wt
        INNER JOIN tags t on t.TgID = wt.WtTgID
        order by TgText
      ) tagssrc
      GROUP BY WtWoID
    ) AS tags on tags.WoID = w.WoID
    LEFT OUTER JOIN wordimages wi on wi.WiWoID = w.WoID
    """

    typecrit = supported_parser_type_criteria()
    wheres = [f"L.LgParserType in ({typecrit})"]

    # Add "where" criteria for all the filters.

    # Have to check for 'null' for language filter.
    # A new user may filter the language when the demo data is loaded,
    # but on "wipe database" the filtLanguage value stored in localdata
    # may be invalid, resulting in the filtLanguage form control actually
    # sending the **string value** "null" here.
    # The other filter values don't change with the data,
    # so we don't need to check for null.
    # Tricky tricky.
    language_id = parameters["filtLanguage"]
    if language_id == "null" or language_id is None:
        language_id = "0"
    language_id = int(language_id)
    if language_id != 0:
        wheres.append(f"L.LgID == {language_id}")

    if parameters["filtParentsOnly"] == "true":
        wheres.append("parents.parentlist IS NULL")

    sql_age_calc = "cast(julianday('now') - julianday(w.wocreated) as int)"
    age_min = parameters["filtAgeMin"].strip()
    if age_min:
        wheres.append(f"{sql_age_calc} >= {int(age_min)}")
    age_max = parameters["filtAgeMax"].strip()
    if age_max:
        wheres.append(f"{sql_age_calc} <= {int(age_max)}")

    status_wheres = ["StID not in (0, 98)"]
    status_min = int(parameters["filtStatusMin"])
    status_max = int(parameters["filtStatusMax"])
    if status_min > 0:
        status_wheres.append(f"StID >= {status_min}")
    if status_max > 0:
        status_wheres.append(f"StID <= {status_max}")
    status_wheres = " AND ".join(status_wheres)
    if parameters["filtIncludeIgnored"] == "true":
        status_wheres = f"({status_wheres} OR StID = 98)"
    wheres.append(status_wheres)

    # Phew.
    return DataTablesSqliteQuery.get_data(
        base_sql + " WHERE " + " AND ".join(wheres), parameters, db.session.connection()
    )



"""
Book domain objects.
"""

from lute.models.book import Book as DBBook, BookTag
from lute.models.language import Language


class Book:  # pylint: disable=too-many-instance-attributes
    """
    A book domain object, to create/edit lute.models.book.Books.

    Book language can be specified either by language_id, or
    language_name.  language_name is useful for loading books via
    scripts/api.  language_id takes precedence.
    """

    def __init__(self):
        self.id = None
        self.language_id = None
        self.language_name = None
        self.title = None
        self.text = None
        self.max_page_tokens = 250
        self.source_uri = None
        self.audio_filename = None
        self.audio_current_pos = None
        self.audio_bookmarks = None
        self.book_tags = []

    def __repr__(self):
        return f"<Book (id={self.id}, title='{self.title}')>"

    def add_tag(self, tag):
        self.book_tags.append(tag)


class Repository:
    """
    Maps Book BO to and from lute.model.Book.
    """

    def __init__(self, _db):
        self.db = _db

    def load(self, book_id):
        "Loads a Book business object for the DBBook."
        dbb = DBBook.find(book_id)
        if dbb is None:
            raise ValueError(f"No book with id {book_id} found")
        return self._build_business_book(dbb)

    def get_book_tags(self):
        "Get all available book tags, helper method."
        bts = self.db.session.query(BookTag).all()
        return [t.text for t in bts]

    def add(self, book):
        """
        Add a book to be saved to the db session.
        Returns DBBook for tests and verification only,
        clients should not change it.
        """
        dbbook = self._build_db_book(book)
        self.db.session.add(dbbook)
        return dbbook

    def delete(self, book):
        """
        Delete.
        """
        if book.id is None:
            raise ValueError(f"book {book.title} not saved")
        b = DBBook.find(book.id)
        self.db.session.delete(b)

    def commit(self):
        """
        Commit everything.
        """
        self.db.session.commit()

    def _build_db_book(self, book):
        "Convert a book business object to a DBBook."

        lang = None
        if book.language_id:
            lang = Language.find(book.language_id)
        elif book.language_name:
            lang = Language.find_by_name(book.language_name)

        b = None
        if book.id is None:
            b = DBBook.create_book(book.title, lang, book.text, book.max_page_tokens)
        else:
            b = DBBook.find(book.id)
        b.title = book.title
        b.source_uri = book.source_uri
        b.audio_filename = book.audio_filename
        b.audio_current_pos = book.audio_current_pos
        b.audio_bookmarks = book.audio_bookmarks

        booktags = []
        for s in book.book_tags:
            booktags.append(BookTag.find_or_create_by_text(s))
        b.remove_all_book_tags()
        for tt in booktags:
            b.add_book_tag(tt)

        return b

    def _build_business_book(self, dbbook):
        "Convert db book to Book."
        b = Book()
        b.id = dbbook.id
        b.language_id = dbbook.language.id
        b.language_name = dbbook.language.name
        b.title = dbbook.title
        b.text = None  # Not returning this for now
        b.source_uri = dbbook.source_uri
        b.audio_filename = dbbook.audio_filename
        b.audio_current_pos = dbbook.audio_current_pos
        b.audio_bookmarks = dbbook.audio_bookmarks
        b.book_tags = [t.text for t in dbbook.book_tags]
        return b



"""
Run pre-recorded smoke tests using playwright.

Notes:

- the db must be reset to the baseline with demo stories
- site must be running (currently hardcoded to port 5000)
- start code gen in another window with `python codegen`,
  then go to http://localhost:5000/ in the new window
- click through etc etc, then stop the code gen, copy-paste
  code here, fix as needed, _then_ shut down

Debugging:

- to debug, can use "page.pause()" to pause the runner.

More notes:

This is _just a smoke test_, it doesn't do any assertions.
The actions were _recorded_ using playwright's supertastic
code generation.  https://playwright.dev/python/docs/codegen

Then I added some tweaks:

Menu sub-items are only visible after hovering over the menu, e.g.:
  page.locator("#menu_books").hover()
  page.locator("#book_new").click()
"""

import os
import time
from playwright.sync_api import Playwright, sync_playwright, expect


def run(p: Playwright) -> None:  # pylint: disable=too-many-statements
    "Run the smoke test."

    # Run headless, can add an env var later for headless or not.
    showbrowser = os.environ.get("SHOW", "") == "true"

    # print(os.environ.get("SHOW"), flush=True)
    # print("-" * 50)
    def _print(s):
        print(s)

    _print("Opening browser.")
    browser = p.chromium.launch(headless=not showbrowser)
    context = browser.new_context()
    context.set_default_timeout(30000)
    page = context.new_page()

    _print("Reset db.")
    page.goto("http://localhost:5000/dev_api/load_demo")

    # Hardcoded port will cause problems ...
    page.goto("http://localhost:5000/")

    # Open Tutorial
    _print("Tutorial check.")
    page.get_by_role("link", name="Tutorial", exact=True).click()
    page.locator("#ID-176-1").click()
    page.frame_locator('iframe[name="wordframe"]').get_by_placeholder(
        "Translation"
    ).click()
    page.frame_locator('iframe[name="wordframe"]').get_by_placeholder(
        "Translation"
    ).fill("big grey thing")
    page.frame_locator('iframe[name="wordframe"]').get_by_role(
        "button", name="Save"
    ).click()
    page.get_by_title(
        "Mark rest as known, mark page as read, then go to next page"
    ).click()
    page.get_by_title("Mark page as read, then go to next page", exact=True).click()
    page.get_by_title("Home").click()

    # Open and archive book.
    _print("Archive.")
    page.get_by_role("link", name="Büyük ağaç").click()
    page.on("dialog", lambda dialog: dialog.accept())
    page.get_by_role("link", name="Archive book").click()

    # Make a new book
    _print("New book.")
    page.locator("#menu_books").hover()
    page.locator("#book_new").click()
    page.locator("#language_id").select_option("4")
    page.get_by_label("Title").click()
    page.get_by_label("Title").fill("Hello")
    page.get_by_label("Text", exact=True).click()
    page.get_by_label("Text", exact=True).fill("Hello there.")
    page.get_by_role("button", name="Save").click()

    # Edit a term.
    _print("Edit term.")
    page.locator("#ID-1-1").click()
    page.frame_locator('iframe[name="wordframe"]').get_by_placeholder(
        "Translation"
    ).click()
    page.frame_locator('iframe[name="wordframe"]').get_by_placeholder(
        "Translation"
    ).fill("Hi.")
    page.frame_locator('iframe[name="wordframe"]').get_by_role(
        "button", name="Save"
    ).click()

    # Archive current book "Hello", check archive.
    _print("Archive.")
    page.get_by_role("link", name="Archive book").click()
    page.locator("#menu_books").hover()
    page.get_by_role("link", name="Book archive").click()
    expect(page.get_by_role("link", name="Hello")).to_be_visible()

    # Open term listing.
    _print("Term listing.")
    page.locator("#menu_terms").hover()
    page.get_by_role("link", name="Terms", exact=True).click()
    page.get_by_role("link", name="Hello").click()
    # TODO testing: restore Sentences smoke test check.
    # page.get_by_role("link", name="Sentences").click()
    page.get_by_role("link", name="Back to list").click()
    # page.pause()

    # TODO issue_336_export_unknown_book_terms: restore this test.
    # _print("Export parent term mapping files.")
    # page.locator("#menu_terms").hover()
    # page.get_by_role("link", name="Parent Term mapping").click()
    # with page.expect_download(timeout=30000) as _:
    #     page.get_by_role("link", name="Tutorial", exact=True).click()

    # Edit language.
    _print("Edit language.")
    page.goto("http://localhost:5000/")
    page.locator("#menu_settings").hover()
    page.get_by_role("link", name="Languages").click()
    page.get_by_role("link", name="English").click()
    page.get_by_role("button", name="Save").click()

    # Wipe the db.
    _print("Reset db.")
    page.get_by_role("link", name="click here").click()

    # Create a new language.
    _print("New language.")
    page.get_by_role("link", name="create your language.").click()
    page.locator("#predefined").select_option("Spanish")
    page.get_by_role("button", name="go").click()
    page.get_by_role("button", name="Save").click()

    # Create a new book for the new lang.
    _print("New book.")
    page.get_by_role("link", name="Create one?").click()
    page.get_by_label("Title").click()
    page.get_by_label("Title").fill("Hola.")
    page.get_by_label("Text", exact=True).fill("Tengo un perro.")
    page.get_by_role("button", name="Save").click()

    # Interact with text.
    page.get_by_text("perro").click()
    page.frame_locator('iframe[name="wordframe"]').get_by_placeholder(
        "Translation"
    ).click()
    page.frame_locator('iframe[name="wordframe"]').get_by_placeholder(
        "Translation"
    ).fill("dog.")
    page.frame_locator('iframe[name="wordframe"]').get_by_role(
        "button", name="Save"
    ).click()

    # Go home, backup is kicked off.
    _print("Disabled: Verify backup started.")
    page.locator("#reading-footer").get_by_role("link", name="Home").click()
    # TODO disabled_backup_check: backup now runs and redirects to home.
    # Not sure how to check it easily ... wait for it to complete.
    time.sleep(4)
    # page.get_by_role("link", name="Back to home.").click()

    # Archive and unarchive.
    # Disabled, the links are now hidden inside a small hover-over dropdown.
    # TODO reactivate_disabled_tests: book links are in a small hover-over list.
    _print("Disabled: Archive and unarchive.")
    ### expect(page.get_by_role("link", name="Hola.")).to_be_visible()
    ### page.get_by_title("Archive", exact=True).click()
    ### expect(page.get_by_role("link", name="Create one?")).to_be_visible()
    ### page.locator("#menu_books").hover()
    ### page.get_by_role("link", name="Book archive").click()
    ### expect(page.get_by_role("link", name="Hola.")).to_be_visible()
    ### page.get_by_title("Unarchive", exact=True).click()
    ### expect(page.get_by_role("link", name="Hola.")).to_be_visible()

    # Import web page.
    _print("Import web page.")
    page.locator("#menu_books").hover()
    page.get_by_role("link", name="Import web page").click()
    page.get_by_label("Import URL").fill(
        "http://localhost:5000/dev_api/fake_story.html"
    )
    page.get_by_role("button", name="Import").click()
    time.sleep(2)
    # Page is imported, form shown, so save it.
    page.get_by_role("button", name="Save").click()
    page.get_by_text("Tengo").click()  # Quick hacky check if exists.
    page.get_by_title("Home").click()
    expect(page.get_by_role("link", name="Mi perro.")).to_be_visible()

    # Check version.
    _print("Version.")
    page.locator("#menu_about").hover()
    page.get_by_role("link", name="Version and software info").click()

    # Custom style.
    _print("Custom style.")
    page.locator("#menu_settings").hover()
    page.get_by_role("link", name="Settings").click()
    page.get_by_label("Custom styles").click()
    page.get_by_label("Custom styles").fill(
        "span.status0 { background-color: yellow; }"
    )
    page.get_by_role("button", name="Save").click()

    # ---------------------
    context.close()
    browser.close()


def test_playwright():
    "Run playwright with tests."
    with sync_playwright() as sp:
        run(sp)



"""
Step defs for term_rendering.feature.
"""
# pylint: disable=missing-function-docstring

from pytest_bdd import given, then, scenarios, parsers

from lute.db import db
from lute.models.language import Language
from lute.term.model import Repository
from lute.read.render.service import get_paragraphs
from lute.read.service import set_unknowns_to_known, bulk_status_update

from tests.utils import add_terms, make_text
from tests.dbasserts import assert_sql_result


# The current language being used.
language = None

# The Text object
text = None

scenarios("rendering.feature")


@given("demo data")
def given_demo_data(app_context):
    "Calling app_context loads the demo data."


@given(parsers.parse("language {langname}"))
def given_lang(langname):
    global language  # pylint: disable=global-statement
    lang = db.session.query(Language).filter(Language.name == langname).first()
    assert lang.name == langname, "sanity check"
    language = lang


@given(parsers.parse("terms:\n{content}"))
def given_terms(content):
    terms = content.split("\n")
    add_terms(language, terms)


@given(parsers.parse('term "{content}" with status {status} and parent "{parenttext}"'))
def given_term_with_status_and_parent(content, status, parenttext):
    r = Repository(db)
    t = r.find_or_new(language.id, content)
    t.status = int(status)
    t.parents.append(parenttext)
    r.add(t)
    r.commit()


@given(parsers.parse('term "{content}" with status {status}'))
def given_term_with_status(content, status):
    r = Repository(db)
    t = r.find_or_new(language.id, content)
    t.status = int(status)
    r.add(t)
    r.commit()


@given(parsers.parse("text:\n{content}"))
def given_text(content):
    global text  # pylint: disable=global-statement
    text = make_text("test", content, language)
    db.session.add(text)
    db.session.commit()


@given("all unknowns are set to known")
def set_to_known():
    set_unknowns_to_known(text)


@given(parsers.parse("bulk status {newstatus} update for terms:\n{terms}"))
def update_status(newstatus, terms):
    bulk_status_update(text, terms.split("\n"), int(newstatus))


def _assert_stringized_equals(stringizer, joiner, expected):
    """
    Get paragraphs and stringize all textitems,
    join and assert equals expected.
    """
    paras = get_paragraphs(text.text, text.book.language)
    ret = []
    for p in paras:
        tis = [t for s in p for t in s.textitems]
        ss = [stringizer(ti) for ti in tis]
        ret.append(joiner.join(ss))
    actual = "/<PARA>/".join(ret)

    expected = expected.split("\n")
    assert actual == "/<PARA>/".join(expected)


@then(parsers.parse("rendered should be:\n{expected}"))
def then_rendered_should_be(expected):
    """
    Renders /term(status)/ /term/ /term/, compares with expected.
    """

    def stringize(ti):
        zws = "\u200B"
        status = ""
        if ti.wo_status not in [None, 0]:
            status = f"({ti.wo_status})"
        return ti.display_text.replace(zws, "") + status

    _assert_stringized_equals(stringize, "/", expected)


@then(parsers.parse("known-only rendered should be:\n{expected}"))
def known_only_rendered_should_be(expected):
    def stringize(ti):
        s = ti.display_text
        if ti.wo_status not in [None, 0]:
            s = f"[[{s}]]"
        zws = "\u200B"
        return s.replace(zws, "")

    _assert_stringized_equals(stringize, "", expected)


@then(parsers.parse("words table should contain:\n{text_lc_content}"))
def then_words_table_contains_WoTextLC(text_lc_content):
    expected = text_lc_content.split("\n")
    sql = "select WoTextLC from words order by WoTextLC"
    assert_sql_result(sql, expected)



"""
Db and image backup.
"""

import os
import re
import shutil
import gzip
from datetime import datetime
import time
from typing import List, Union

from lute.db import db
from lute.models.setting import SystemSetting
from lute.models.book import Book
from lute.models.term import Term


class BackupException(Exception):
    """
    Raised if export bombs for some reason.
    """


class DatabaseBackupFile:
    """
    A representation of a lute backup file to hold metadata attributes.
    """

    def __init__(self, filepath: Union[str, os.PathLike]):
        if not os.path.exists(filepath):
            raise BackupException(f"No backup file at {filepath}.")

        name = os.path.basename(filepath)
        if not re.match(r"(manual_)?lute_backup_", name):
            raise BackupException(f"Not a valid lute database backup at {filepath}.")

        self.filepath = filepath
        self.name = name
        self.is_manual = self.name.startswith("manual_")

    def __lt__(self, other):
        return self.last_modified < other.last_modified

    @property
    def last_modified(self) -> datetime:
        return datetime.fromtimestamp(os.path.getmtime(self.filepath)).astimezone()

    @property
    def size_bytes(self) -> int:
        return os.path.getsize(self.filepath)

    @property
    def size(self) -> str:
        """
        A human-readable string representation of the size of the file.

        Eg.
        1746 bytes
        4 kB
        27 MB
        """
        s = self.size_bytes
        if s >= 1e9:
            return f"{round(s * 1e-9)} GB"
        if s >= 1e6:
            return f"{round(s * 1e-6)} MB"
        if s >= 1e3:
            return f"{round(s * 1e-3)} KB"
        return f"{s} bytes"


def create_backup(app_config, settings, is_manual=False, suffix=None):
    """
    Create backup using current app config, settings.

    is_manual is True if this is a user-triggered manual
    backup, otherwise is False.

    suffix can be specified for test.

    settings are from Setting.get_backup_settings().
      - backup_enabled
      - backup_dir
      - backup_auto
      - backup_warn
      - backup_count
      - last_backup_datetime
    """
    if not os.path.exists(settings.backup_dir):
        raise BackupException("Missing directory " + settings.backup_dir)

    ### Timing helper for when implement audio backup.
    # def _print_now(msg):
    #     now = datetime.now().strftime("%H-%M-%S")
    #     print(f"{now} - {msg}", flush=True)

    _mirror_images_dir(app_config.userimagespath, settings.backup_dir)

    prefix = "manual_" if is_manual else ""
    if suffix is None:
        suffix = datetime.now().strftime("%Y-%m-%d_%H%M%S")
    fname = f"{prefix}lute_backup_{suffix}.db"
    backupfile = os.path.join(settings.backup_dir, fname)

    f = _create_db_backup(app_config.dbfilename, backupfile)
    _remove_excess_backups(settings.backup_count, settings.backup_dir)
    return f


def should_run_auto_backup(backup_settings):
    """
    True (if applicable) if last backup was old.
    """
    bs = backup_settings
    if bs.backup_enabled is False or bs.backup_auto is False:
        return False

    last = bs.last_backup_datetime
    if last is None:
        return True

    curr = int(time.time())
    diff = curr - last
    return diff > 24 * 60 * 60


def backup_warning(backup_settings):
    "Get warning if needed."
    if not backup_settings.backup_warn:
        return ""

    have_books = db.session.query(db.session.query(Book).exists()).scalar()
    have_terms = db.session.query(db.session.query(Term).exists()).scalar()
    if have_books is False and have_terms is False:
        return ""

    last = backup_settings.last_backup_datetime
    if last is None:
        return "Never backed up."

    curr = int(time.time())
    diff = curr - last
    old_backup_msg = "Last backup was more than 1 week ago."
    if diff > 7 * 24 * 60 * 60:
        return old_backup_msg

    return ""


def _create_db_backup(dbfilename, backupfile):
    "Make a backup."
    shutil.copy(dbfilename, backupfile)
    f = f"{backupfile}.gz"
    with open(backupfile, "rb") as in_file, gzip.open(
        f, "wb", compresslevel=4
    ) as out_file:
        shutil.copyfileobj(in_file, out_file)
    os.remove(backupfile)
    SystemSetting.set_last_backup_datetime(int(time.time()))
    return f


def skip_this_backup():
    "Set the last backup time to today."
    SystemSetting.set_last_backup_datetime(int(time.time()))


def _remove_excess_backups(count, outdir):
    "Remove old backups."
    files = [f for f in list_backups(outdir) if not f.is_manual]
    files.sort(reverse=True)
    to_remove = files[count:]
    for f in to_remove:
        os.remove(f.filepath)


def _mirror_images_dir(userimagespath, outdir):
    "Copy the images to backup."
    target_dir = os.path.join(outdir, "userimages_backup")
    target_dir = os.path.abspath(target_dir)
    if not os.path.exists(target_dir):
        os.mkdir(target_dir)
    shutil.copytree(userimagespath, target_dir, dirs_exist_ok=True)


def list_backups(outdir) -> List[DatabaseBackupFile]:
    "List all backup files."
    return [
        DatabaseBackupFile(os.path.join(outdir, f))
        for f in os.listdir(outdir)
        if re.match(r"(manual_)?lute_backup_", f)
    ]



"""
Mapping parents.
"""

# TODO issue_336_export_unknown_book_terms: this blueprint can be killed when 336 is done.

import os
from flask import Blueprint, current_app, render_template, send_file
from sqlalchemy import asc
from lute.db import db
from lute.models.book import Book
from lute.models.language import Language
from lute.term_parent_map.service import export_unknown_terms


bp = Blueprint("term_parent_map", __name__, url_prefix="/term_parent_map")


@bp.route("/index", methods=["GET", "POST"])
def index():
    """
    Show books and languages
    """

    # sqlalchemy _requires_ "== False" for the comparison!
    # pylint: disable=singleton-comparison
    books = (
        db.session.query(Book)
        .filter(Book.archived == False)
        .order_by(asc(Book.title))
        .all()
    )
    languages = db.session.query(Language).order_by(asc(Language.name)).all()
    return render_template(
        "/term_parent_map/index.html", books=books, languages=languages
    )


# TODO issue_336_export_unknown_book_terms: move this route, or something like it, to book actions.
@bp.route("/export_book/<int:bookid>", methods=["GET"])
def export_book(bookid):
    "Generate a file and return it."
    outfile = os.path.join(current_app.env_config.temppath, "export_book.txt")
    book = db.session.get(Book, bookid)
    export_unknown_terms(book, outfile)
    return send_file(outfile, as_attachment=True, download_name="unknown_terms.txt")



"""
DB Backup tests.
"""

import os
from datetime import datetime, timezone
from unittest.mock import Mock, patch
import pytest

from lute.backup.service import (
    create_backup,
    BackupException,
    should_run_auto_backup,
    backup_warning,
    DatabaseBackupFile,
    list_backups,
)
from lute.models.setting import BackupSettings

# pylint: disable=missing-function-docstring
# Test method names are pretty descriptive already.

utc = timezone.utc


@pytest.fixture(name="bkp_dir")
def fixture_backup_directory(testconfig):
    "Create clean backup dir."
    base_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../"))
    backup_dir = os.path.join(base_dir, "zz_bkp")
    if not os.path.exists(backup_dir):
        os.mkdir(backup_dir)

    # Cleanup prior to test.
    cleanup_directory(backup_dir)
    cleanup_directory(testconfig.userimagespath)

    yield backup_dir

    # Don't clean up after the test, in case want to check manually.
    # cleanup_directories(backup_dir, image_dir)


def cleanup_directory(directory):
    "Delete all files and subfolders."
    for root, dirs, files in os.walk(directory, topdown=False):
        for f in files:
            os.remove(os.path.join(root, f))
        for d in dirs:
            os.rmdir(os.path.join(root, d))


@pytest.fixture(name="backup_settings")
def fixture_backup_settings(app_context, bkp_dir):
    # app_context is passed so that the db session is available.
    ret = BackupSettings.get_backup_settings()
    ret.backup_dir = bkp_dir
    ret.backup_enabled = True
    yield ret


def _mock_backup_file(directory, name, time, size):
    file_mock = Mock()
    file_mock.name = name
    file_mock.path = os.path.join(directory, name)
    file_mock.size = size
    with open(file_mock.path, "wb") as f:
        f.seek(file_mock.size - 1)
        f.write(b"\0")
    os.utime(file_mock.path, (time.timestamp(), time.timestamp()))
    return file_mock


@pytest.fixture(name="auto_backup_file")
def fixture_auto_backup_file(bkp_dir):
    yield _mock_backup_file(
        bkp_dir,
        "lute_backup_2024-01-01-000000.db.gz",
        datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc),
        567890,
    )


@pytest.fixture(name="manual_backup_file")
def fixture_manual_backup_file(bkp_dir):
    yield _mock_backup_file(
        bkp_dir,
        "manual_lute_backup_2024-02-01-000000.db.gz",
        datetime(2024, 2, 1, 0, 0, 0, tzinfo=utc),
        123450,
    )


def test_backup_writes_file_to_output_dir(testconfig, bkp_dir, backup_settings):
    file_content = b"imagefile"
    img1 = os.path.join(testconfig.userimagespath, "1")
    if not os.path.exists(img1):
        os.mkdir(img1)
    with open(os.path.join(img1, "file.txt"), "wb") as f:
        f.write(file_content)

    create_backup(testconfig, backup_settings)
    assert len(os.listdir(bkp_dir)) == 2  # db and directory
    assert len(os.listdir(os.path.join(bkp_dir, "userimages_backup", "1"))) == 1
    assert os.path.exists(os.path.join(bkp_dir, "userimages_backup", "1", "file.txt"))


def test_timestamp_added_to_db_name(testconfig, bkp_dir, backup_settings):
    assert os.listdir(bkp_dir) == [], "empty dirs at start"
    create_backup(testconfig, backup_settings)
    dbfile = [f for f in os.listdir(bkp_dir) if f.startswith("lute_backup_2")]
    assert len(dbfile) == 1, "db found"


def test_backup_fails_if_missing_output_dir(testconfig, backup_settings):
    backup_settings.backup_dir = "some_missing_dir"
    with pytest.raises(BackupException, match="Missing directory some_missing_dir"):
        create_backup(testconfig, backup_settings)


def test_user_can_configure_rolling_backup_count(testconfig, bkp_dir, backup_settings):
    backup_settings.backup_count = 2
    for i in range(1, 10):
        create_backup(testconfig, backup_settings, suffix=f"0{i}")
    expected_files = [f"lute_backup_0{i}.db.gz" for i in range(8, 10)]
    db_files = [f for f in os.listdir(bkp_dir) if f.endswith(".gz")]
    assert db_files.sort() == expected_files.sort()


def test_all_manual_backups_are_kept(testconfig, bkp_dir, backup_settings):
    backup_settings.backup_count = 2
    for i in range(1, 10):
        create_backup(testconfig, backup_settings, suffix=f"0{i}", is_manual=True)
    expected_files = [f"lute_backup_0{i}.db.gz" for i in range(1, 10)]
    db_files = [f for f in os.listdir(bkp_dir) if f.endswith(".gz")]
    assert db_files.sort() == expected_files.sort()


def test_last_import_setting_is_updated_on_successful_backup(
    testconfig, backup_settings
):
    assert backup_settings.last_backup_datetime is None, "no backup"
    create_backup(testconfig, backup_settings)
    updated = BackupSettings.get_backup_settings()
    assert updated.last_backup_datetime is not None, "set"


def test_should_not_run_autobackup_if_auto_is_no_or_false(backup_settings):
    backup_settings.backup_enabled = True
    backup_settings.backup_auto = False
    assert should_run_auto_backup(backup_settings) is False


def test_autobackup_returns_true_if_never_backed_up(backup_settings):
    backup_settings.backup_enabled = True
    backup_settings.backup_auto = True
    backup_settings.last_backup_datetime = None
    assert should_run_auto_backup(backup_settings) is True


def test_autobackup_returns_true_if_last_backed_up_over_one_day_ago(backup_settings):
    backup_settings.backup_auto = True
    # backup_settings.backup_warn = True
    curr_datetime = datetime.now()
    one_day_ago = curr_datetime.timestamp() - 24 * 60 * 60

    backup_settings.last_backup_datetime = one_day_ago - 10
    assert should_run_auto_backup(backup_settings) is True

    backup_settings.last_backup_datetime = one_day_ago + 10
    assert should_run_auto_backup(backup_settings) is False


def test_warn_if_last_backup_never_happened_or_is_old(backup_settings):
    curr_datetime = datetime.now()
    one_week_ago = curr_datetime.timestamp() - 7 * 24 * 60 * 60

    backup_settings.backup_warn = True
    backup_settings.last_backup_datetime = None
    assert backup_warning(backup_settings) == "Never backed up."

    backup_settings.last_backup_datetime = one_week_ago + 10
    assert backup_warning(backup_settings) == ""

    backup_settings.last_backup_datetime = one_week_ago - 10
    assert backup_warning(backup_settings) == "Last backup was more than 1 week ago."

    backup_settings.backup_warn = False
    assert backup_warning(backup_settings) == ""


def test_database_backup_file_on_nonexistent_path(bkp_dir):
    file = os.path.join(bkp_dir, "lute_backup_nonexistent.db.gz")
    with pytest.raises(BackupException) as excinfo:
        DatabaseBackupFile(file)
    assert str(excinfo.value) == f"No backup file at {file}."


def test_database_backup_file_with_non_lute_backup(bkp_dir):
    file = os.path.join(bkp_dir, "some_other_file.db.gz")
    with open(file, "wb") as f:
        f.write(b"\0")
    with pytest.raises(BackupException) as excinfo:
        DatabaseBackupFile(file)
    assert str(excinfo.value) == f"Not a valid lute database backup at {file}."


def test_database_backup_file_with_auto_backup_returns_success(auto_backup_file):
    dbf = DatabaseBackupFile(auto_backup_file.path)

    assert auto_backup_file.name == dbf.name
    assert auto_backup_file.path == dbf.filepath
    assert auto_backup_file.size == dbf.size_bytes
    assert dbf.size == "568 KB"
    assert dbf.last_modified == datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc)


def test_database_backup_file_for_auto_is_not_manual(auto_backup_file):
    dbf = DatabaseBackupFile(auto_backup_file.path)
    assert not dbf.is_manual


def test_database_backup_file_for_manual_is_not_manual(manual_backup_file):
    dbf = DatabaseBackupFile(manual_backup_file.path)
    assert dbf.is_manual


def test_database_backup_file_size_formatting(auto_backup_file):
    dbf = DatabaseBackupFile(auto_backup_file.path)

    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 11234567890):
        assert dbf.size == "11 GB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 1123456789):
        assert dbf.size == "1 GB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 19876543):
        assert dbf.size == "20 MB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 7654321):
        assert dbf.size == "8 MB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 1024):
        assert dbf.size == "1 KB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 1001):
        assert dbf.size == "1 KB"
    with patch("lute.backup.service.DatabaseBackupFile.size_bytes", 944):
        assert dbf.size == "944 bytes"


def test_list_backups_returns_both(
    bkp_dir, auto_backup_file, manual_backup_file
):  # pylint: disable=unused-argument
    backups = list_backups(bkp_dir)
    assert len(backups) == 2
    assert all(isinstance(backup, DatabaseBackupFile) for backup in backups)
    assert any(not backup.is_manual for backup in backups)
    assert any(backup.is_manual for backup in backups)


def test_backup_listing_sorts_by_modified(
    bkp_dir, auto_backup_file, manual_backup_file
):  # pylint: disable=unused-argument
    backups = list_backups(bkp_dir)
    assert len(backups) == 2

    backups.sort()
    assert backups[0].last_modified == datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc)
    assert backups[1].last_modified == datetime(2024, 2, 1, 0, 0, 0, tzinfo=utc)

    backups.sort(reverse=True)
    assert backups[0].last_modified == datetime(2024, 2, 1, 0, 0, 0, tzinfo=utc)
    assert backups[1].last_modified == datetime(2024, 1, 1, 0, 0, 0, tzinfo=utc)



"""
Tasks to run using Invoke.

Ref https://docs.pyinvoke.org/en/stable/index.html

Samples:

invoke lint
invoke coverage --html

invoke --list          # list all tasks
invoke --help <cmd>    # See docstrings and help notes
"""

import os
import sys
import subprocess
from datetime import datetime
import requests
from invoke import task, Collection
from lute.config.app_config import AppConfig


@task
def lint(c):
    "Run pylint on lute/ and tests/."
    # Formats: https://pylint.pycqa.org/en/latest/user_guide/usage/output.html
    msgfmt = "--msg-template='{path} ({line:03d}): {msg} ({msg_id} {symbol})'"
    c.run(f"pylint {msgfmt} tasks.py lute/ tests/")


@task
def todos(c):
    """
    Print code TODOs.
    """
    c.run("python utils/todos.py")


@task(help={"port": "optional port to run on; default = 5000"})
def start(c, port=5000):
    """
    Start the dev server, using script dev.py.
    """
    c.run(f"python -m devstart --port {port}")


@task
def search(c, search_for):
    """
    Search the code for a string.
    """
    thisdir = os.path.dirname(os.path.realpath(__file__))
    devscript = os.path.join(thisdir, "utils", "findstring.sh")
    c.run(f'{devscript} "{search_for}"')


@task
def _ensure_test_db(c):  # pylint: disable=unused-argument
    "Quits if not a test db. (Hidden task)"
    ac = AppConfig(AppConfig.default_config_filename())
    if ac.is_test_db is False:
        print(
            f"""
        QUITTING TASK FOR NON-TEST DB.
        Your database name must start with test_ to run this task.
        Your config.yml file has dbname = {ac.dbname}.
        """
        )
        sys.exit(1)


@task(pre=[_ensure_test_db])
def test(c):
    """
    Unit tests only.
    """
    c.run("pytest --ignore=./tests/acceptance")


def _site_is_running(useport=None):
    """
    Return true if site is running on port, or default 5000.
    """
    if useport is None:
        useport = 5000

    url = f"http://localhost:{useport}"
    try:
        print(f"checking for site at {url} ...")
        resp = requests.get(url, timeout=5)
        if resp.status_code != 200:
            raise RuntimeError(f"Got code {resp.status_code} ... ???")
        print("Site running, using that for tests.")
        print()
        return True
    except requests.exceptions.ConnectionError:
        print(f"URL {url} not reachable, will start new server at that port.")
        print()
        return False


@task(
    pre=[_ensure_test_db],
    help={
        "port": "optional port to run on; creates server if needed.",
        "show": "print data",
        "noheadless": "run as non-headless (default is headless, i.e. not shown)",
        "kflag": "optional -k flag argument",
        "exitfirst": "exit on first failure",
        "verbose": "make verbose",
    },
)
def accept(  # pylint: disable=too-many-arguments
    c,
    port=5000,
    show=False,
    noheadless=False,
    kflag=None,
    exitfirst=False,
    verbose=False,
):
    """
    Start lute, run tests/acceptance tests, screenshot fails.

    If no port specified, use default 5000.

    If Lute's not running on specified port, start a server.
    """
    run_test = [
        "pytest",
        "tests/acceptance",
        "--splinter-screenshot-dir=tests/acceptance/failure_screenshots",
        "--splinter-webdriver=chrome",
        f"--port={port}",
    ]

    if show:
        run_test.append("-s")
    if not noheadless:
        run_test.append("--headless")
    if kflag:
        run_test.append("-k")
        run_test.append(kflag)
    if exitfirst:
        run_test.append("--exitfirst")
    if verbose:
        run_test.append("-vv")

    tests_failed = False
    if _site_is_running(port):
        c.run(" ".join(run_test))
    else:
        cmd = ["python", "-m", "tests.acceptance.start_acceptance_app", f"{port}"]
        with subprocess.Popen(cmd) as app_process:
            try:
                subprocess.run(run_test, check=True)
            except subprocess.CalledProcessError:
                # This just means a test failed.  We don't need to see
                # a stack trace, the assert failures are already displayed.
                tests_failed = True
            finally:
                app_process.terminate()

    if tests_failed:
        raise RuntimeError("tests failed")


@task(pre=[_ensure_test_db])
def playwright(c):
    """
    Start lute, run playwright tests.  export SHOW=true env var to run non-headless.

    Only uses port 5000.

    If Lute's not running on specified port, start a server.
    """
    run_test = ["pytest", "tests/playwright/playwright.py", "-s"]

    tests_failed = False
    port = 5000
    if _site_is_running(port):
        c.run(" ".join(run_test))
    else:
        cmd = ["python", "-m", "tests.acceptance.start_acceptance_app", f"{port}"]
        with subprocess.Popen(cmd) as app_process:
            try:
                subprocess.run(run_test, check=True)
            except subprocess.CalledProcessError:
                # This just means a test failed.  We don't need to see
                # a stack trace, the assert failures are already displayed.
                tests_failed = True
            finally:
                app_process.terminate()

    if tests_failed:
        raise RuntimeError("tests failed")


@task(pre=[_ensure_test_db], help={"html": "open html report"})
def coverage(c, html=False):
    """
    Run coverage (using non-acceptance tests only), open report if needed.

    Running tests including the acceptance tests is slow,
    and doesn't affect the coverage stats enough to justify it.
    """
    c.run("coverage run -m pytest tests/")
    if html:
        c.run('coverage html --omit="tests/*"')
        c.run("open htmlcov/index.html")
    else:
        cmd = 'coverage report --sort=cover --show-missing --omit="tests/*"'
        c.run(cmd)


@task(post=[lint])
def black(c):
    "black-format things."
    c.run("python -m black .")


@task(pre=[test, accept, playwright])
def fulltest(c):  # pylint: disable=unused-argument
    """
    Run full tests check.
    """
    print("Done.")


@task(pre=[fulltest, black, lint])
def full(c):  # pylint: disable=unused-argument
    """
    Run full check and lint.
    """
    print("Done.")


ns = Collection()
ns.add_task(fulltest)
ns.add_task(full)
ns.add_task(lint)
ns.add_task(test)
ns.add_task(accept)
ns.add_task(playwright)
ns.add_task(coverage)
ns.add_task(todos)
ns.add_task(start)
ns.add_task(search)
ns.add_task(black)


##############################
# DB tasks


@task(pre=[_ensure_test_db])
def db_wipe(c):
    """
    Wipe the data from the testing db; factory reset settings. :-)

    Can only be run on a testing db.
    """
    c.run("pytest -m dbwipe")
    print("ok")


@task(pre=[_ensure_test_db])
def db_reset(c):
    """
    Reset the database to the demo data.

    Can only be run on a testing db.
    """
    c.run("pytest -m dbdemoload")
    print("ok")


def _schema_dir():
    "Return full path to schema dir."
    thisdir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(thisdir, "lute", "db", "schema")


def _do_schema_export(c, destfile, header_notes, taskname):
    """
    Generate the dumpfile at destfile.
    """
    destfile = os.path.join(_schema_dir(), destfile)
    tempfile = f"{destfile}.temp"
    commands = f"""
    echo "-- ------------------------------------------" > {tempfile}
    echo "-- {header_notes}" >> {tempfile}
    echo "-- Migrations tracked in _migrations, settings reset." >> {tempfile}
    echo "-- Generated from 'inv {taskname}'" >> {tempfile}
    echo "-- ------------------------------------------" >> {tempfile}
    echo "" >> {tempfile}
    sqlite3 data/test_lute.db .dump >> {tempfile}
    """
    c.run(commands)

    os.rename(tempfile, destfile)
    print(f"{destfile} updated (git diff follows):")
    print("DIFF START " + "-" * 38)
    # -U0 = no context before or after changes.
    c.run(f"git diff -U0 -- {destfile}")
    print("DIFF END " + "-" * 40)
    print()


@task
def db_export_baseline(c):
    """
    Reset the db, and create a new baseline db file from the current db.
    """

    # Running the delete task before this one as a pre- step was
    # causing problems (sqlite file not in correct state), so this
    # asks the user to verify.
    text = input("Have you reset the db?  (y/n): ")
    if text != "y":
        print("quitting.")
        return
    _do_schema_export(
        c, "baseline.sql", "Baseline db with demo data.", "db.export.baseline"
    )

    fname = os.path.join(_schema_dir(), "baseline.sql")
    print(f"Verifying {fname}")
    with open(fname, "r", encoding="utf-8") as f:
        checkstring = "Tutorial follow-up"
        if checkstring in f.read():
            print(f'"{checkstring}" found, likely ok.')
        else:
            print(f'"{checkstring}" NOT FOUND, SOMETHING LIKELY WRONG.')
            raise RuntimeError(f'Missing "{checkstring}" in exported file.')


@task
def db_export_empty(c):
    """
    Create a new empty db file from the current db.

    This assumes that the current db is in data/test_lute.db.
    """

    # Running the delete task before this one as a pre- step was
    # causing problems (sqlite file not in correct state), so this
    # asks the user to verify.
    text = input("Have you **WIPED** the db?  (y/n): ")
    if text != "y":
        print("quitting.")
        return
    _do_schema_export(c, "empty.sql", "EMPTY DB.", "db.export.empty")


@task(help={"suffix": "suffix to add to filename."})
def db_newscript(c, suffix):  # pylint: disable=unused-argument
    """
    Create a new migration, <datetime>_suffix.sql
    """
    now = datetime.now()
    fnow = now.strftime("%Y%m%d")
    filename = f"{fnow}_{suffix}.sql"
    destfile = os.path.join(_schema_dir(), "migrations", filename)
    with open(destfile, "w", encoding="utf-8") as f:
        f.write("-- TODO - fill this in.")
    print("migration created:")
    print(destfile)


dbtasks = Collection("db")
dbtasks.add_task(db_reset, "reset")
dbtasks.add_task(db_wipe, "wipe")
dbtasks.add_task(db_newscript, "newscript")
dbexport = Collection("export")
dbexport.add_task(db_export_baseline, "baseline")
dbexport.add_task(db_export_empty, "empty")
dbtasks.add_collection(dbexport)

ns.add_collection(dbtasks)



"""
Parser registry.

List of available parsers.
"""

from lute.parse.base import AbstractParser
from lute.parse.space_delimited_parser import SpaceDelimitedParser, TurkishParser
from lute.parse.mecab_parser import JapaneseParser
from lute.parse.character_parser import ClassicalChineseParser


# List of ALL parsers available, not necessarily all supported.
# This design feels fishy, but it suffices for now.
parsers = {
    "spacedel": SpaceDelimitedParser,
    "turkish": TurkishParser,
    "japanese": JapaneseParser,
    "classicalchinese": ClassicalChineseParser,
}


def _supported_parsers():
    "Get the supported parsers."
    ret = {}
    for k, v in parsers.items():
        if v.is_supported():
            ret[k] = v
    return ret


def get_parser(parser_name) -> AbstractParser:
    "Return the supported parser with the given name."
    if parser_name in _supported_parsers():
        pclass = parsers[parser_name]
        return pclass()
    raise ValueError(f"Unknown parser type '{parser_name}'")


def is_supported(parser_name) -> bool:
    "Return True if the specified parser is supported, false otherwise or if not found."
    if parser_name not in parsers:
        return False
    p = parsers[parser_name]
    return p.is_supported()


def supported_parsers():
    """
    Dictionary of supported parser strings and class names, for UI.

    For select list entries, use supported_parsers().items().
    """
    ret = []
    for k, v in _supported_parsers().items():
        ret.append([k, v.name()])
    return ret


def supported_parser_types():
    """
    List of supported Language.parser_types
    """
    return list(_supported_parsers().keys())



"""
Text tests.
"""

from datetime import datetime
from lute.models.book import Book, Text


def transform_sentence(s):
    zws = chr(0x200B)  # zero-width space
    return s.text_content.replace(zws, "/")


def test_sentence_lifecycle(english):
    """
    Sentences must be generated when a Text is saved with the ReadDate saved.
    Sentences are only used for reference lookups.
    """
    b = Book("hola", english)
    t = Text(b, "Tienes un perro. Un gato.")

    assert len(t.sentences) == 0, "no sentences"

    t.read_date = datetime.now()
    assert len(t.sentences) == 2, "sentences are created when read_date is set"

    assert (
        transform_sentence(t.sentences[0]) == "/Tienes/ /un/ /perro/./"
    ), "1st sentence"
    assert transform_sentence(t.sentences[1]) == "/Un/ /gato/./", "2"

    t.text = "Tengo un coche."
    assert len(t.sentences) == 1, "changed"

    assert transform_sentence(t.sentences[0]) == "/Tengo/ /un/ /coche/./", "changed"



"""
Create todo report.
"""

import os
import re
import subprocess


class Todout:
    "Gather and categorize todos."

    def __init__(self):
        self.verbose = False

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, v):
        self._verbose = v

    def debug_print(self, msg):
        if self.verbose:
            print(msg)

    def find_files(self, directory, exclude_dirs=None, exclude_files=None, maxdepth=10):
        "Find files matching pattern."
        if not os.path.exists(directory) or not os.path.isdir(directory):
            raise RuntimeError(f"Missing directory: {directory}")

        self.debug_print(
            f"Searching {directory} excluding dirs {exclude_dirs} and files {exclude_files}"
        )

        if exclude_dirs is None:
            exclude_dirs = []
        if exclude_files is None:
            exclude_files = []

        exclude_files = [
            e if (e.startswith("./") or e.startswith("/")) else f"./{e}"
            for e in exclude_files
        ]
        files = []

        os.chdir(directory)
        self.debug_print(f"Searching {os.getcwd()}")
        self.debug_print(f"excluding dirs {exclude_dirs}")

        cmd = f"find . -type f -maxdepth {maxdepth} -print0 | xargs -0 grep -li todo 2>/dev/null"
        result = subprocess.run(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )

        files = result.stdout.split("\n")
        # self.debug_print('initial files:')
        # self.debug_print(files)

        files = [
            f
            for f in files
            if not any(f.startswith(e) or f.startswith(f"{e}/") for e in exclude_dirs)
        ]

        files = [f for f in files if f not in exclude_files]

        if self.verbose:
            print("Found files:\n" + "\n".join(files))

        return files

    def grepfiles(self, root_dir, files):
        "Search."
        results = []
        for f in files:
            fullname = os.path.abspath(os.path.join(root_dir, f))
            command = f"grep -i todo {fullname}"
            # self.debug_print(command)
            result = subprocess.run(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=True,
            )
            result_lines = result.stdout.split("\n")
            self.debug_print(result_lines)
            result_lines = [r for r in result_lines if not r.startswith("Binary file")]
            results.extend([(f, r.strip()) for r in result_lines if r.strip() != ""])
        return results

    def get_grouping(self, line):
        "Get the 'todo group' (TODO <group>:) for the current line."
        tmp = line.strip()
        tmp = re.sub("^.*todo", "", tmp)
        if tmp.startswith(":"):
            return ""
        if ":" not in tmp:
            return ""
        return tmp.split(":", 1)[0].strip()

    def get_todo_data(
        self, directory, exclude_dirs=None, exclude_files=None, maxdepth=10
    ):
        """
        Get the files, gather todo data.
        """
        files = self.find_files(directory, exclude_dirs, exclude_files, maxdepth)
        files = [f for f in files if f is not None and f.strip() != ""]
        ret = []
        for f in files:
            results = self.grepfiles(directory, [f])
            for f, line in results:
                ret.append(
                    {
                        "file": f,
                        "line": line,
                        "group": self.get_grouping(line.lower()),
                    }
                )
        return ret


def write_report(data):
    """
    Print summary to console.
    """
    data = [r for r in data if r["file"] != ""]
    # print(data)
    groups = sorted(set(d["group"] for d in data))

    for g in groups:
        curr_group_data = [gd for gd in data if gd["group"] == g]

        heading = g if g != "" else "<None>"
        print("\nGroup: " + heading)
        for gd in curr_group_data:
            print(f"  {gd['file'].ljust(50)}:  {gd['line']}")
    print("\n")


t = Todout()
t.verbose = False
current_script_path = os.path.abspath(__file__)
rootdir = os.path.join(os.path.dirname(current_script_path), "..")
rootdir = os.path.abspath(rootdir)
venvdir = os.path.join(rootdir, ".venv")
htmlcovdir = os.path.join(rootdir, "htmlcov")
lutedir = os.path.join(rootdir, "lute")
testdir = os.path.join(rootdir, "tests")

d = t.get_todo_data(
    rootdir,
    [venvdir, lutedir, testdir, "./htmlcov", "./.git", "./.venv"],
    ["./utils/todos.py", "./.pylintrc", "./tasks.py", "./docs/development.md"],
    99,
)
write_report(d)



"""
Book mapping checks.
"""

from datetime import datetime
import pytest
from lute.models.book import Book, BookTag
from lute.book.stats import BookStats
from lute.db import db
from tests.dbasserts import assert_sql_result, assert_record_count_equals


@pytest.fixture(name="simple_book")
def fixture_simple_book(english):
    "Single page book with some associated objects."
    b = Book.create_book("hi", english, "some text")
    b.texts[0].read_date = datetime.now()
    bt = BookTag.make_book_tag("hola")
    b.book_tags.append(bt)
    return b


def test_save_book(empty_db, simple_book):
    """
    Check book mappings.
    """
    b = simple_book
    db.session.add(b)
    db.session.commit()

    sql = "select BkID, BkTitle, BkLgID from books"
    assert_sql_result(sql, ["1; hi; 1"], "book")

    sql = "select TxID, TxBkID, TxText from texts"
    assert_sql_result(sql, ["1; 1; some text"], "texts")

    sql = "select * from sentences"
    assert_sql_result(sql, ["1; 1; 1; /some/ /text/"], "sentences")

    sql = "select * from booktags"
    assert_sql_result(sql, ["1; 1"], "booktags")

    sql = "select * from tags2"
    assert_sql_result(sql, ["1; hola; "], "tags2")


def test_delete_book(empty_db, simple_book):
    """
    Most tables should clear out after delete.
    """
    b = simple_book
    db.session.add(b)
    db.session.commit()

    db.session.delete(b)
    db.session.commit()

    for t in ["books", "texts", "sentences", "booktags"]:
        assert_record_count_equals(t, 0, f"{t} deleted")

    sql = "select * from tags2"
    assert_sql_result(sql, ["1; hola; "], "tags2 remain")


def test_save_and_delete_created_book(english):
    """
    Verify book orm mappings.
    """
    content = "Some text here. Some more text"
    b = Book.create_book("test", english, content, 3)
    db.session.add(b)
    db.session.commit()
    sql = f"select TxOrder, TxText from texts where TxBkID = {b.id}"
    expected = ["1; Some text here.", "2; Some more text"]
    assert_sql_result(sql, expected, "texts")

    db.session.delete(b)
    db.session.commit()
    assert_sql_result(sql, [], "texts deleted")


def test_load_book_loads_lang(empty_db, simple_book):
    """
    Check book mappings.
    """
    b = simple_book
    db.session.add(b)
    db.session.commit()

    findbook = db.session.query(Book).filter(Book.title == "hi").first()
    assert findbook.title == "hi", "title"
    assert findbook.language.name == "English", "check lang"

    for b in db.session.query(Book).all():
        assert b.language is not None, "have lang object"

    books_to_update = (
        db.session.query(Book)
        .filter(~Book.id.in_(db.session.query(BookStats.BkID)))
        .all()
    )
    for b in books_to_update:
        assert b.language is not None, "have lang object"



"""
Functions to manage demo database.

Lute db comes pre-loaded with some demo data.  User can view Tutorial,
wipe data, etc.

The db settings table contains a record, StKey = 'IsDemoData', if the
data is demo.
"""

from sqlalchemy import text
import lute.language.service
from lute.book.model import Repository
from lute.book.stats import refresh_stats
from lute.models.setting import SystemSetting
from lute.db import db
import lute.db.management


def _demo_languages():
    """
    Demo languages to be loaded for new users.
    Also loaded during tests.
    """
    return [
        "Arabic",
        "Classical Chinese",
        "Czech",
        "English",
        "French",
        "German",
        "Greek",
        "Hindi",
        "Japanese",
        "Russian",
        "Sanskrit",
        "Spanish",
        "Turkish",
    ]


def contains_demo_data():
    """
    True if IsDemoData setting is present.
    """
    ss = SystemSetting.get_value("IsDemoData")
    if ss is None:
        return False
    return True


def remove_flag():
    """
    Remove IsDemoData setting.
    """
    if not contains_demo_data():
        raise RuntimeError("Can't delete non-demo data.")

    SystemSetting.delete_key("IsDemoData")
    db.session.commit()


def tutorial_book_id():
    """
    Return the book id of the tutorial.
    """
    if not contains_demo_data():
        return None
    sql = """select BkID from books
    inner join languages on LgID = BkLgID
    where LgName = 'English' and BkTitle = 'Tutorial'
    """
    r = db.session.execute(text(sql)).first()
    if r is None:
        return None
    return int(r[0])


def delete_demo_data():
    """
    If this is a demo, wipe everything.
    """
    if not contains_demo_data():
        raise RuntimeError("Can't delete non-demo data.")
    remove_flag()
    lute.db.management.delete_all_data()


# Loading demo data.


def load_demo_languages():
    """
    Load selected predefined languages.  Assume everything is supported.

    This method will also be called during acceptance tests, so it's public.
    """
    demo_langs = _demo_languages()
    langs = [
        lute.language.service.get_language_def(langname)["language"]
        for langname in demo_langs
    ]
    supported = [lang for lang in langs if lang.is_supported]
    for lang in supported:
        db.session.add(lang)
    db.session.commit()


def load_demo_stories():
    "Load the stories."
    demo_langs = _demo_languages()
    langdefs = [
        lute.language.service.get_language_def(langname) for langname in demo_langs
    ]
    langdefs = [d for d in langdefs if d["language"].is_supported]

    r = Repository(db)
    for d in langdefs:
        for b in d["books"]:
            r.add(b)
    r.commit()

    SystemSetting.set_value("IsDemoData", True)
    db.session.commit()
    refresh_stats()


def load_demo_data():
    """
    Load the data.
    """
    load_demo_languages()
    load_demo_stories()
    SystemSetting.set_value("IsDemoData", True)
    db.session.commit()
